/////////////////////////////////////////////////////////////////////////////
// Name:        TASApp.cpp
// Purpose:     Defines the main application class
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "stdafx.h"
#include "wxprec.h"
#include <wx/datetime.h>
#include <wx/splash.h>
#include <wx/image.h>
#include <wx/xrc/xmlres.h>
#include <wx/file.h>
#include <wx/tokenzr.h>
#include "TASSimFrm.h"
#include "TASApp.h"
#include "resource.h"
#include "TAS.h"
#include "TASSimFrm.h"
#include "TASEditor.h"
#include "TASDockers.h"
#include "TASTiles.h"
#include "TASTilePanel.h"
#include "TASDebug.h"

#ifdef __WXMAC__
/*void wxOnAssert(char const*, int, char const*, char const*, char const*)
{
}*/

/*void wxStringBase::InitWith(char const* c, size_t s1, size_t s2)
{

}*/
#endif

TASApp*      g_pTASApp = NULL;
TASContext*  g_pContext = NULL;

extern void TASInitXmlResource(); // defined in generated file

TASContext::TASContext()
              : wxConfig(_T("TAS")),
                m_setSimWin(TASSimWinName, wxPoint(20, 20), wxSize(700, 700), true, false),
                m_setEditorWin(TASEditorWinName, wxPoint(320, 20), wxSize(700, 700), true, false)
{
  m_iTASDimMode               = TAS_2D_MODE;      //  Default to 2D mode
  m_iTASSimMode               = TAS_ATAM_MODE;    //  Default to aTAM simulation
  m_iTASOptMode               = TAS_OPT_MODE_1;   //  Default to optimized
  m_iTASOptModeThreshold      = 2000;             //  Try 2000 for now...
  m_maskSimTileDisplay        = TAS_MASK_DISPALL;
  m_maskEditorTileDisplay     = TAS_MASK_DISPALL;
  m_bUseEditor                = true;
  m_bOutputDebugMsgs          = false;
  m_bCheckForLocalDeterminism = false;
  m_iTileAdditionsPerStep     = TAS_TILE_ADD_ONE;
  m_bAsyncDisplay             = false;
  m_bUse2HAMHack              = false;
  m_iNum2HAMThreads           = wxThread::GetCPUCount();
  m_iTASTileConcMode          = TAS_TILECONC_MODE_OFF;
  m_refreshRate               = 1;
  m_bTASBatchMode             = false;
  m_lCurrNumRuns              = 0;
  m_bDoReset                  = false;

  wxString sBatchSettingsName = _T("batch_mode_settings");
  m_settingsBatchMode.SetName(sBatchSettingsName);

  wxString sOutputBatchModeName = _T("ISU_TAS_simulation_output");
  m_outputBatchMode.SetName(sOutputBatchModeName);
}

void TASContext::LoadSavedSettings()
{
  //  Get saved settings for dimension mode
#ifdef TAS_BUILD_3D
  m_iTASDimMode = Read(TASDimModeName, (long)TAS_2D_MODE);
#else
  m_iTASDimMode = TAS_2D_MODE;
#endif  // TAS_BUILD_3D

  m_iTASSimMode = Read(TASSimModeName, (long)TAS_ATAM_MODE);

  m_fnTilesetFile       = Read(_T("tdsFile"),  DefaultTilesetFileName);
  m_fnTilespaceFile     = Read(_T("tdpFile"),  DefaultTilespaceFileName);

  m_colSpaceBkgrnd.Set(Read(_T("bgColor"),   _T("#F3F77D")));//"#000000"));
  m_colDefaultTile.Set(Read(_T("tileColor"), _T("#FFFFFF")));
  m_colDefaultText.Set(Read(_T("textColor"), _T("#000000")));

  m_maskSimTileDisplay    = Read(_T("simTileDisplayMask"),    TAS_MASK_DISPALL);
  m_maskEditorTileDisplay = Read(_T("editorTileDisplayMask"), TAS_MASK_DISPALL);

/*
  m_sCellValuePaneInfo    = Read(_T("cellValPanePersp"),  _T(""));
  m_sSpaceConfigPaneInfo  = Read(_T("spacePanePersp"),    _T(""));
  m_sAxesPaneInfo         = Read(_T("axesPanePersp"),     _T(""));
  m_sOverviewPaneInfo     = Read("overviewPanePersp",     _T(""));
*/

  m_sSimFramePersp          = Read(_T("simFramePersp"),     _T(""));
  m_sEditorFramePersp       = Read(_T("editorFramePersp"),    _T(""));

  m_setSimWin.LoadSettings(this);
  m_setEditorWin.LoadSettings(this);

  Read(TASUseEditor, &m_bUseEditor, true);
  Read(TASOutputDebugMsgs, &m_bOutputDebugMsgs, false);
  Read(TASCheckLocalDeterminism, &m_bCheckForLocalDeterminism, false);
  Read(TASNumTileAdditions,&m_iTileAdditionsPerStep, TAS_TILE_ADD_ONE);
  Read(TASUse2HAMHack,&m_bUse2HAMHack, false);
  Read(TASNum2HAMThreads, &m_iNum2HAMThreads, wxThread::GetCPUCount());

  m_bDoSave = true;
}

void TASContext::SaveCurrentSettings()
{
  if (m_fnTilesetFile.IsRelative()) {
    m_fnTilesetFile.MakeAbsolute();
  }
  if (m_fnTilespaceFile.IsRelative()) {
    m_fnTilespaceFile.MakeAbsolute();
  }
  Write(_T("tdsFile"),  m_fnTilesetFile.GetFullPath());
  Write(_T("tdpFile"),  m_fnTilespaceFile.GetFullPath());
  Write(TASDimModeName, m_iTASDimMode);
  Write(TASSimModeName, m_iTASSimMode);

  Write(_T("bgColor"), m_colSpaceBkgrnd.GetAsString(wxC2S_HTML_SYNTAX));
  Write(_T("tileColor"), m_colDefaultTile.GetAsString(wxC2S_HTML_SYNTAX));
  Write(_T("textColor"), m_colDefaultText.GetAsString(wxC2S_HTML_SYNTAX));

  Write(_T("simTileDisplayMask"),     (long)m_maskSimTileDisplay);
  Write(_T("editorTileDisplayMask"),  (long)m_maskEditorTileDisplay);

  m_setSimWin.SaveCurrentSettings(this);
  m_setEditorWin.SaveCurrentSettings(this);

/*
  //  Set best sizes to be the current sizes, then get pane info for each dockable window
  wxRect r1 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin).BestSize(r1.width, r1.height);
  m_sCellValuePaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin));

  wxRect r3 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin).BestSize(r3.width, r3.height);
  m_sSpaceConfigPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin));

  wxRect r4 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin).BestSize(r4.width, r4.height);
  m_sAxesPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin));

  wxRect r5 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin).BestSize(r5.width, r5.height);
  m_sOverviewPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin));

  Write(_T("cellValPanePersp"), m_sCellValuePaneInfo);
  Write(_T("spacePanePersp"),   m_sSpaceConfigPaneInfo);
  Write(_T("axesPanePersp"),    m_sAxesPaneInfo);
  Write(_T("overviewPanePersp"),m_sOverviewPaneInfo);
*/

  m_sSimFramePersp        = g_pTASApp->m_pSimFrame->m_mgr.SavePerspective();
  Write(_T("simFramePersp"),   m_sSimFramePersp);

  if (g_pTASApp->m_pEditorFrame) {
    m_sEditorFramePersp       = g_pTASApp->m_pEditorFrame->m_mgr.SavePerspective();
    Write(_T("editorFramePersp"),   m_sEditorFramePersp);
  }

  Write(TASUseEditor, m_bUseEditor);
  Write(TASOutputDebugMsgs, m_bOutputDebugMsgs);
  Write(TASCheckLocalDeterminism, m_bCheckForLocalDeterminism);
  Write(TASNumTileAdditions,m_iTileAdditionsPerStep);
  Write(TASUse2HAMHack,m_bUse2HAMHack);
  Write(TASNum2HAMThreads, m_iNum2HAMThreads);
}

bool TASContext::IsAssemblyFileOK()
{
  if ((m_fnTilespaceFile.GetFullName().Len() > 0) &&
      (m_fnTilespaceFile.GetFullName() != DefaultTilespaceFileName) &&
      (m_fnTilespaceFile.IsOk()))
  {
    return true;
  }
  return false;
}

bool TASContext::IsTilesetFileOK()
{
  if ((m_fnTilesetFile.GetFullName().Len() > 0) &&
      (m_fnTilesetFile.GetFullName() != DefaultTilesetFileName) &&
      (m_fnTilesetFile.IsOk()))
  {
    return true;
  }
  return false;
}

TASWinSettings::TASWinSettings(wxString strName)
{
  m_pWin        = NULL;
  m_strName     = strName;
  m_ptLoc       = wxDefaultPosition;
  m_szSize      = wxDefaultSize;
  m_bVisible    = true;
  m_bMaximized  = false;
}

TASWinSettings::TASWinSettings(wxString strName, wxPoint ptLoc, wxSize szSize, bool bVisible, bool bMaximized)
{
  m_pWin        = NULL;
  m_strName     = strName;
  m_ptLoc       = ptLoc;
  m_szSize      = szSize;
  m_bVisible    = bVisible;
  m_bMaximized  = bMaximized;
}

TASWinSettings& TASWinSettings::operator=(const TASWinSettings& rhs)
{
  m_strName     = rhs.m_strName;
  m_ptLoc       = rhs.m_ptLoc;
  m_szSize      = rhs.m_szSize;
  m_bVisible    = rhs.m_bVisible;
  m_bMaximized  = rhs.m_bMaximized;

  return *this;
}

void TASWinSettings::GetCurrentSettings()
{
  if (m_pWin) {
    if (m_pWin->IsShown()) {
      m_bVisible = true;
    } else {
      m_bVisible = false;
    }

    m_bMaximized = m_pWin->IsMaximized();
    if (m_bMaximized) {
      m_pWin->Maximize(false);
    }

    m_ptLoc.x = m_pWin->GetPosition().x;
    m_ptLoc.y = m_pWin->GetPosition().y;
    m_szSize.SetWidth(m_pWin->GetSize().GetWidth());
    m_szSize.SetHeight(m_pWin->GetSize().GetHeight());

  } else {
    m_bVisible = false;
  }
}

void TASWinSettings::SaveCurrentSettings(wxConfig* pConfig)
{
  GetCurrentSettings();

  pConfig->Write(m_strName + _T("LocX"), m_ptLoc.x);
  pConfig->Write(m_strName + _T("LocY"), m_ptLoc.y);
  pConfig->Write(m_strName + _T("SizeX"), m_szSize.GetWidth());
  pConfig->Write(m_strName + _T("SizeY"), m_szSize.GetHeight());
  pConfig->Write(m_strName + _T("Vis"), (m_bVisible ? 1 : 0));
  pConfig->Write(m_strName + _T("Maxd"), (m_bMaximized ? 1 : 0));
}

void TASWinSettings::LoadSettings(wxConfig* pConfig)
{
  m_ptLoc.x = pConfig->Read(m_strName + _T("LocX"), m_ptLoc.x);
  m_ptLoc.y = pConfig->Read(m_strName + _T("LocY"), m_ptLoc.y);
  m_szSize.SetWidth(pConfig->Read(m_strName + _T("SizeX"), m_szSize.GetWidth()));
  m_szSize.SetHeight(pConfig->Read(m_strName + _T("SizeY"), m_szSize.GetHeight()));
  m_bVisible = ((pConfig->Read(m_strName + _T("Vis"), (long)1) == 0) ? false : true);
  m_bMaximized = ((pConfig->Read(m_strName + _T("Maxd"), (long)0) == 0) ? false : true);
}


// Create a new application object: this macro will allow wxWindows to create
// the application object during program execution (it's better than using a
// static object for many reasons) and also declares the accessor function
// wxGetApp() which will return the reference of the right type (i.e. TASApp and
// not wxApp)
IMPLEMENT_APP(TASApp)

// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// the application class
// ----------------------------------------------------------------------------

static wxCriticalSection* s_critSect_logMsg = NULL;

wxMutex* g_pMutexTileSpaceGuard = NULL;

extern TASTileType* g_pEmptyTileType;
extern wxCriticalSection* g_pCritSectShouldDie;
extern wxCriticalSection* g_pCritSectDeleteList;

extern wxFont* g_pFontLabel;
extern wxFont* g_pFontLabelSVG;
extern wxFont* g_pFontGlues;
extern wxFont* g_pFontGluesSVG;
extern wxFont* g_pFontGlues3D;

void InitGlobals()
{
  g_pEmptyTileType = new TASTileType();
  g_pCritSectShouldDie = new wxCriticalSection();
  g_pCritSectDeleteList = new wxCriticalSection();

  g_pFontLabel    = new wxFont(12, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false);
  g_pFontLabelSVG = new wxFont(6, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false);
  g_pFontGlues    = new wxFont(8, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);
  g_pFontGluesSVG = new wxFont(4, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);
  g_pFontGlues3D  = new wxFont(12, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);
}

void DeleteGlobals()
{
  if (g_pMutexTileSpaceGuard) delete g_pMutexTileSpaceGuard;
  if (s_critSect_logMsg) delete s_critSect_logMsg;
  if (g_pEmptyTileType) delete g_pEmptyTileType;
  if (g_pCritSectShouldDie) delete g_pCritSectShouldDie;
  if (g_pCritSectDeleteList) delete g_pCritSectDeleteList;

  if (g_pFontLabel) delete g_pFontLabel;
  if (g_pFontLabelSVG) delete g_pFontLabelSVG;
  if (g_pFontGlues) delete g_pFontGlues;
  if (g_pFontGluesSVG) delete g_pFontGluesSVG;
  if (g_pFontGlues3D) delete g_pFontGlues3D;
}

TASApp::TASApp()
{
  //  Create the mutex that will be used to protect against concurrent accesses
  //  to the tile space by multiple threads
  g_pMutexTileSpaceGuard   = new wxMutex();

  m_pSimFrame       = NULL;
  m_pEditorFrame    = NULL;

  Srand(0, true);
  m_randVal         = m_randSeed;

  g_pTASApp         = this;
}

TASApp::~TASApp()
{
  DeleteGlobals();
}

void TASApp::SetTASDimMode(int mode)
{
  g_pContext->m_iTASDimMode = mode;
}

int  TASApp::GetTASDimMode()
{
  return g_pContext->m_iTASDimMode;
}

void TASApp::SetTASSimMode(int mode)
{
  g_pContext->m_iTASSimMode = mode;
}

int  TASApp::GetTASSimMode()
{
  return g_pContext->m_iTASSimMode;
}

// 'Main program' equivalent: the program execution "starts" here
bool TASApp::OnInit()
{
  InitGlobals();

  // Resource initialization
  wxImage::AddHandler(new wxGIFHandler);
  wxImage::AddHandler(new wxICOHandler);
  wxImage::AddHandler(new wxXPMHandler);
  wxXmlResource::Get()->InitAllHandlers();
  TASInitXmlResource();

  g_pContext = new TASContext();
  wxString sCrashCheck;
  g_pContext->Read(TASCrashCheck, &sCrashCheck);
  if (!sCrashCheck.IsEmpty()) {
    int ret = wxMessageBox(wxT("It appears that ISU TAS did not exit correctly the last time it was run.  This could possibly be due to corrupt or invalid settings.  Please press 'Yes' to delete the currently saved settings, or 'No' to continue using the current settings"), wxT("ISU TAS Error Recovery"), wxYES_NO|wxICON_ERROR);
    if (ret == wxYES) {
      g_pContext->DeleteAll();
    }
  }
  g_pContext->Write(TASCrashCheck, wxT("Running"));

  //  Create the global tileset and tile space objects
  NewTAS();

  g_pContext->LoadSavedSettings();

  ProcessCommandLineFiles(argc, argv);

//MJP TODO: if tdp is set but not tds, look for tds spec in tdp


  // create the main TAS window
  m_pSimFrame = new TASSimulatorFrame(ID_TAS_SIMULATOR_FRAME, this, _("TAS 2.0"), g_pContext->m_setSimWin.m_ptLoc, 
                                      g_pContext->m_setSimWin.m_szSize, wxDEFAULT_FRAME_STYLE);
  m_pSimFrame->Maximize(g_pContext->m_setSimWin.m_bMaximized);
  g_pContext->m_setSimWin.m_pWin = m_pSimFrame;

  if (g_pContext->m_bUseEditor) {
    CreateEditor();
  }

  m_pSimFrame->Show(TRUE);
  SetTopWindow(m_pSimFrame);

  m_pSimFrame->Refresh();
  m_pSimFrame->Update();

///////MJP: stuff to wrap up into a separate function
  if (g_pContext->IsTilesetFileOK()) {
    if (ReloadTilesetIntoEditor()) {
      if (LoadTilesetIntoSimulator() &&
          g_pContext->IsAssemblyFileOK())
      {
        ReloadAssembly();
      }
    }
  }

  m_pSimFrame->m_panel->SetViewableSpace(TASCoords(0, 0, 0), TASCoords(0, 0, 0), TAS_3D_VIEWBY_UNLIMITED);
  if (m_pSimFrame->m_pCellValueWin) {
    m_pSimFrame->m_pCellValueWin->SetFields();
  }
  if (m_pSimFrame->m_pSpaceConfigWin) {
    m_pSimFrame->m_pSpaceConfigWin->SetViewType(TAS_3D_VIEWBY_UNLIMITED);
    m_pSimFrame->m_pSpaceConfigWin->SetCurrentDims();
  }

///////////

  if (GetTASDimMode() == TAS_2D_MODE) {
    ((TASTilePanel*)m_pSimFrame->m_panel)->CheckCanvasSize();
    m_pSimFrame->m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
    m_pSimFrame->m_panel->UpdateSpace(true);
//    m_pEditorFrame->ProcessUpdatedTileset();
  }

//MJP: an unworking hack to try to fix the fact that the 3-D canvas doesn't always fill the panel
//m_pSimFrame->SetTASDimMode(TAS_3D_MODE, true);

  // success: wxApp::OnRun() will be called which will enter the main message
  // loop and the application will run. If we returned FALSE here, the
  // application would exit immediately.
  return TRUE;
}

bool TASApp::CreateEditor()
{
  // create the editor window
  m_pEditorFrame = new TASEditorFrame(ID_TAS_EDITOR_FRAME, this, g_pContext->m_setEditorWin.m_ptLoc, g_pContext->m_setEditorWin.m_szSize);
  m_pEditorFrame->Maximize(g_pContext->m_setEditorWin.m_bMaximized);
  g_pContext->m_setEditorWin.m_pWin = m_pEditorFrame;

  m_pEditorFrame->Show(TRUE);

  return true;
}

size_t GetTileTypeCount(const wxString& str)
{
  size_t count  = 0;
  size_t pos    = 0;
  while (pos < str.length()) {
    size_t loc = str.find(_T("TILENAME "), pos);
    if (loc != std::string::npos) {
      count++;
      pos = loc + wxStrlen(_T("TILENAME "));
    } else {
      break;
    }
  }
  return count;
}

wxString GetTdsFileNameFromTdpFile(wxString& sTASFileName)
{
  wxString sRet;

  wxFFile f(sTASFileName);
  if (f.IsOpened()) {
    wxString sContents;
    if (f.ReadAll(&sContents)) {
      if (sContents.Find(wxT(".tds")) != wxNOT_FOUND) {
        sRet = sContents.SubString(0, sContents.Find(wxT(".tds"))-1) + wxT(".tds");
      }
    }
    f.Close();
  }
  return sRet;
}

bool TASApp::LoadTAS(wxString& sTASFileName)
{
  if (!::wxFileExists(sTASFileName)) {
    wxString errMsg = wxString::Format(_T("Cannot open file:\n'%s'.\nPlease make sure the file exists and permissions allow this."), sTASFileName.c_str());
    wxMessageBox(errMsg, _T("Error opening tile assembly system"), wxICON_ERROR|wxOK);
    return false;
  }

  wxString sTdsFileName = GetTdsFileNameFromTdpFile(sTASFileName);
  bool bTdsOk = false;
  if (sTdsFileName.length() > 0) {
    wxFileName fnTds(sTdsFileName);
    wxFileName fnTdp(sTASFileName);
    if (fnTds.IsRelative()) {               //  First look for the .tds file in the same place as the .tdp if not fully specified
      fnTds.Assign(fnTdp.GetPath(), fnTds.GetName(), fnTds.GetExt());
    }
    if (::wxFileExists(fnTds.GetFullPath())) {
      sTdsFileName = fnTds.GetFullPath();
      bTdsOk = true;
    }
    if (!bTdsOk) {
      if (::wxFileExists(sTdsFileName)) {   //  Now possibly look for it in the current directory
        bTdsOk = true;
      }
    }
  }

  if (!bTdsOk) {
    sTdsFileName = sTASFileName;
    sTdsFileName.Replace(_T(".tdp"), _T(".tds"));
    if (!::wxFileExists(sTdsFileName)) {
      wxString errMsg = wxString::Format(_T("Unable to find a corresponding '.tds' file. Please make sure one exists."));
      wxMessageBox(errMsg, _T("Error opening tile assembly system"), wxICON_ERROR|wxOK);
      return false;
    } else {
      int useFile = wxMessageBox(wxString::Format(_T("A designated tile set file was not found.  Do you want to use the file\n'%s'?\n(If so, to avoid this message in the future, please re-save the assembly\nto link it with this file.)"), sTdsFileName.c_str()),
                                 _T("Confirm tile set file"), 
                                 wxYES_NO);

      if (useFile != wxYES) {
        wxString errMsg = wxString::Format(_T("Unable to find a corresponding '.tds' file. Please make sure one exists."));
        wxMessageBox(errMsg, _T("Error opening tile assembly system"), wxICON_ERROR|wxOK);
        return false;
      }
    }
  }

  wxFileName fnTds(sTdsFileName);
  wxFileName fnTdp(sTASFileName);
  if (fnTds.IsRelative()) {
    fnTds.Assign(fnTdp.GetPath(), fnTds.GetName(), fnTds.GetExt());
  }
  sTdsFileName = fnTds.GetFullPath();
  wxString sTdpFileName = fnTdp.GetFullPath();
  return (LoadTilesetIntoEditor(sTdsFileName) && LoadTilesetIntoSimulator() && LoadAssembly(sTdpFileName));
}

void TASApp::SetDimModeFromTileset()
{
  int mode = TAS_2D_MODE;
  for (size_t i=0; i<g_pEditorTileset->size(); i++) {
    if ((g_pEditorTileset->at(i)->m_sides[TAS_UP].m_iStrength != 0) ||
        (g_pEditorTileset->at(i)->m_sides[TAS_DOWN].m_iStrength != 0))
      {
      mode = TAS_3D_MODE;
      break;
    }
  }
  if (g_pContext->m_iTASDimMode != mode) {
    if (m_pSimFrame) {
      m_pSimFrame->SetTASDimMode(mode, true);
    }
    if (m_pEditorFrame) {
      m_pEditorFrame->SetDimMode();
      g_pEditorTileset->ReIndex();
      g_pTileSpace->m_tileset.ReIndex();
    }
  }
}

bool TASApp::LoadTilesetIntoEditor(wxString& sTilesetFileName)
{
  wxFFile f(sTilesetFileName, _T("r"));
  if (f.IsOpened()) {
    wxString sContents;
    if (f.ReadAll(&sContents)) {
      size_t ttCount = GetTileTypeCount(sContents);

      wxString sMsg;
      sMsg << ttCount << _T(" tiletypes being loaded into editor...\n");
      SetEditorOutputText(sMsg, false, false);

      if (ttCount > g_pContext->m_iTASOptModeThreshold) {
        g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

          wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
          SetEditorOutputText(sMsg2, true, false);

      } else {
        g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
      }
      *g_pEditorTileset = sContents;
      f.Close();
      g_pContext->m_fnTilesetFile = sTilesetFileName;
      SetDimModeFromTileset();
      if (m_pEditorFrame) {
        m_pEditorFrame->ProcessUpdatedTileset();
      }
      return true;
    }
    f.Close();
  }
  return false;
}

bool TASApp::LoadTilesetFromSimulator()
{
  wxString sMsg;
  sMsg << g_pTileSpace->m_tileset.size() << _T(" tiletypes being loaded into editor...\n");
  SetEditorOutputText(sMsg, false, false);

  if (g_pTileSpace->m_tileset.size() > g_pContext->m_iTASOptModeThreshold) {
    g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

    wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
    SetEditorOutputText(sMsg2, true, false);

  } else {
    g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
  }
  ((TASTileTypeSet&)*g_pEditorTileset) = g_pTileSpace->m_tileset;
  if (m_pEditorFrame) {
    m_pEditorFrame->ProcessUpdatedTileset();
  }
  g_pEditorTileset->SetSimMod(false);
  return true;
}

bool TASApp::LoadTilesetIntoSimulator()
{
  wxString sMsg;
  sMsg << g_pEditorTileset->size() << _T(" tiletypes being loaded into simulator...\n"); 
  SetSimOutputText(sMsg, false, false);

  if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
    wxString sMsg2 = wxString::Format(_T("This tileset is very large, so some seed assembly editing enhancements are disabled.\n")); 
    SetSimOutputText(sMsg2, true, false);
  }
  g_pTileSpace->m_tileset = ((TASTileTypeSet&)*g_pEditorTileset);
  g_pEditorTileset->SetSimMod(false);
  m_pSimFrame->ProcessUpdatedTileset();
  return true;
}

bool TASApp::LoadAssembly(wxString& sAssemblyFileName)
{
  wxFFile f(sAssemblyFileName, _T("r"));
  if (f.IsOpened()) {
    SetTASSimMode(GetTASSimModeFromFile(&f));
    CreateSpace();
    bool bRet = g_pTileSpace->LoadSpaceDefinitionFromFile(&f);
    if (bRet) {
      g_pContext->m_fnTilespaceFile = sAssemblyFileName;
      if (m_pSimFrame) {
        m_pSimFrame->SetTASSimMode(g_pContext->m_iTASSimMode);
        m_pSimFrame->ProcessUpdatedTilespace();
        m_pSimFrame->Refresh();
      }
    }
    f.Close();
    return bRet;
  }
  return false;
}

int TASApp::GetTASSimModeFromFile(wxFFile* f)
{
  wxString sFileContents;
  if (f->IsOpened()) {
    if (f->ReadAll(&sFileContents)) {
      f->Seek(0);
      wxString sModePrefix(_T("Mode="));
      wxString sLine = sFileContents.BeforeFirst('\n');
      sFileContents = sFileContents.AfterFirst('\n');
      while (sFileContents.Len() > 0 || sLine.Len() > 0) {
        wxStringTokenizer tok(sLine);
        if ((tok.CountTokens() == 1) && (sLine.StartsWith(sModePrefix))) {
          wxString sMode = sLine.SubString(sModePrefix.Len(), sLine.Len()).Strip();
          if (sMode.CmpNoCase(TAS_MODE_2HAM_STR) == 0) {
            return TAS_2HAM_MODE;
          } else if(sMode.CmpNoCase(TAS_MODE_KTAM_STR) == 0) {
			return TAS_KTAM_MODE;
		  } else {
            return TAS_ATAM_MODE;
          }
        }
        sLine = sFileContents.BeforeFirst(_T('\n'));
        sFileContents = sFileContents.AfterFirst(_T('\n'));
      }
    }
  }
  return TAS_ATAM_MODE;
}

bool TASApp::LoadStageDefinitions(wxString& sStageFileName)
{
  sStageFileName.Trim().Trim(false);
  if (sStageFileName.Length() == 0) {
    g_pTileSpace->m_bUseStages = false;
    g_pTileSpace->m_vectStages.clear();
    g_pContext->m_fnStageFile.Clear();

    wxString sMsg = wxString::Format(_T("Not using a stage definitions file.\n")); 
    SetSimOutputText(sMsg, false, false);

    return true;
  } else {
    wxFileName fname = sStageFileName;
    if (fname.IsRelative()) {
      wxString path = g_pContext->m_fnTilespaceFile.GetPath();
      fname.SetPath(path);
      sStageFileName = fname.GetFullPath();
    }

    wxFFile f(sStageFileName, _T("r"));
    if (f.IsOpened()) {
      bool bRet = g_pTileSpace->LoadStageDefinitionsFromFile(&f);
      if (bRet) {
        g_pContext->m_fnStageFile = sStageFileName;
        wxString sMsg = wxString::Format(_T("Loaded stage definitions file '%s'\n"), sStageFileName.c_str()); 
        SetSimOutputText(sMsg, false, false);
        ResetAssembly();
        if (m_pSimFrame) {
          m_pSimFrame->DoRefresh();
          if (m_pSimFrame->m_panel) {
            m_pSimFrame->m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
          }
        }
      }
      f.Close();
      return bRet;
    }
    return false;
  }
}

bool TASApp::ReloadTAS()
{
  return (ReloadTilesetIntoEditor() && LoadTilesetIntoSimulator() && ReloadAssembly());
}

bool TASApp::ReloadTilesetIntoEditor()
{
  if (g_pContext->IsTilesetFileOK()) {
    wxFFile f;
    if (f.Open(g_pContext->m_fnTilesetFile.GetFullPath())) {
      wxString sContents;
      if (f.ReadAll(&sContents)) {
        size_t ttCount = GetTileTypeCount(sContents);

        wxString sMsg = wxString::Format(_T("%u tiletypes being loaded into editor...\n"), ttCount);
        SetEditorOutputText(sMsg, false, false);

        if (ttCount > g_pContext->m_iTASOptModeThreshold) {
          g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

          wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
          SetEditorOutputText(sMsg2, true, false);

        } else {
          g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
        }
        *g_pEditorTileset = sContents;
        f.Close();
        SetDimModeFromTileset();
        if (m_pEditorFrame) {
          m_pEditorFrame->ProcessUpdatedTileset();
        }
        return true;
      }
      f.Close();
    }
  }
  return false;
}

bool TASApp::CreateSpace()
{
    if (g_pTileSpaceBackup) {
    g_pTileSpaceBackup->ResetSpace();
  }
  if (g_p2HAMTileSpace) {
    g_p2HAMTileSpace->ResetSpace();
  }

  if (g_pContext->m_iTASSimMode == TAS_ATAM_MODE) {
    if (!g_pTileSpaceBackup) {
      g_pTileSpaceBackup = new TASTileSpace(g_pContext->m_iTASDimMode == TAS_2D_MODE ? 2 : 3);
    }
    if ((g_pTileSpace) && (g_pTileSpaceBackup != g_pTileSpace)) {
      g_pTileSpaceBackup->m_tileset = g_pTileSpace->m_tileset;
    }
    g_pTileSpace = g_pTileSpaceBackup;
  } else if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
    if (!g_p2HAMTileSpace) {
      g_p2HAMTileSpace = new TAS2HAMTileSpace(g_pContext->m_iTASDimMode == TAS_2D_MODE ? 2 : 3);
    }
    if ((g_pTileSpace) && (g_p2HAMTileSpace != g_pTileSpace)) {
      g_p2HAMTileSpace->m_tileset = g_pTileSpace->m_tileset;
    }
    g_pTileSpace = g_p2HAMTileSpace;
  } else if (g_pContext->m_iTASSimMode == TAS_KTAM_MODE) {
    if (!g_pkTAMTileSpace) {
      g_pkTAMTileSpace = new TASkTAMTileSpace(g_pContext->m_iTASDimMode == TAS_2D_MODE ? 2 : 3);
    }
    if ((g_pTileSpace) && (g_pkTAMTileSpace != g_pTileSpace)) {
      g_pkTAMTileSpace->m_tileset = g_pTileSpace->m_tileset;
    }
    g_pTileSpace = g_pkTAMTileSpace;
  }
  return true;
}

bool TASApp::ReloadAssembly()
{
  if (g_pContext->IsAssemblyFileOK()) {
    wxFFile f;
    if (f.Open(g_pContext->m_fnTilespaceFile.GetFullPath())) {
      SetTASSimMode(GetTASSimModeFromFile(&f));
      CreateSpace();
      bool bRet = g_pTileSpace->LoadSpaceDefinitionFromFile(&f);
      if (bRet) {
        if (m_pSimFrame) {
          m_pSimFrame->SetTASSimMode(g_pContext->m_iTASSimMode);
          m_pSimFrame->ProcessUpdatedTilespace();
        }
      }
      f.Close();
      return bRet;
    }
  }
  return false;
}

bool TASApp::ResetTAS()
{
  return (/*LoadTilesetIntoEditor() &&*/ResetAssembly());
}

bool TASApp::ResetAssembly()
{
  g_pTileSpace->ResetSpace(true);
  if (m_pSimFrame) {
    m_pSimFrame->ProcessUpdatedTilespace();
  }
  return true;
}

bool TASApp::SaveTASAs(wxString& sTASFileName)
{
  if (SaveAssemblyAs(sTASFileName)) {
    wxString sTds = sTASFileName;
    sTds.Replace(_T(".tdp"), _T(".tds"));
    return SaveTilesetAs(sTds);
  }
  return false;
}

bool TASApp::SaveTilesetAs(wxString& sTilesetFileName)
{
  wxFFile f(sTilesetFileName, _T("wb"));
  if (f.IsOpened()) {
    g_pContext->m_fnTilesetFile = sTilesetFileName;
    bool bRet = g_pEditorTileset->SaveToFile(&f);
    f.Close();
    return bRet;
  } else {
    wxString errMsg = wxString::Format(_T("There has been an error saving the tileset to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), sTilesetFileName.c_str());
    wxMessageBox(errMsg, _T("Error saving tileset"), wxICON_ERROR|wxOK);
  }
  return false;
}

bool TASApp::SaveAssemblyAs(wxString& sAssmblyFileName)
{
  wxFFile f(sAssmblyFileName, _T("wb"));
  if (f.IsOpened()) {
    g_pContext->m_fnTilespaceFile = sAssmblyFileName;
    m_pSimFrame->SetTitle();
    bool bRet = g_pTileSpace->SaveToFile(&f);
    f.Close();
    return bRet;
  } else {
    wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), sAssmblyFileName.c_str());
    wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
  }
  return false;
}

bool TASApp::SaveTAS()
{
  return (SaveTileset() && SaveAssembly());
}

bool TASApp::SaveTileset()
{
  if (g_pContext->IsTilesetFileOK()) {
    wxFFile f(g_pContext->m_fnTilesetFile.GetFullPath(), _T("wb"));
    if (f.IsOpened()) {
      bool bRet = g_pEditorTileset->SaveToFile(&f);
      f.Close();
      return bRet;
    } else {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile set to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), g_pContext->m_fnTilesetFile.GetFullPath().c_str());
      wxMessageBox(errMsg, _T("Error saving tile set"), wxICON_ERROR|wxOK);
      return false;
    }
  } else {
    return false;
  }
}

bool TASApp::SaveAssembly()
{
  if (g_pContext->IsAssemblyFileOK()) {
    wxFFile f(g_pContext->m_fnTilespaceFile.GetFullPath(), _T("wb"));
    if (f.IsOpened()) {
      bool bRet = g_pTileSpace->SaveToFile(&f);
      f.Close();
      return bRet;
    } else {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), g_pContext->m_fnTilespaceFile.GetFullPath().c_str());
      wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
      return false;
    }
  } else {
    return false;
  }
}

bool TASApp::NewTAS()
{
  return (NewTileset() && NewAssembly());
}

bool TASApp::NewTileset()
{
  g_pContext->m_fnTilesetFile = DefaultTilesetFileName;
  if (g_pEditorTileset) {
    delete g_pEditorTileset;
  }
  g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
  g_pEditorTileset = new TASTileTypeSetEx();
  if (m_pEditorFrame) {
    m_pEditorFrame->ProcessUpdatedTileset();
  }
  return true;
}

bool TASApp::NewAssembly()
{
  g_pContext->m_fnTilespaceFile = DefaultTilespaceFileName;
  CreateSpace();
  if (m_pSimFrame) {
    m_pSimFrame->ProcessUpdatedTilespace();
  }
  return true;
}

void TASApp::SetTemp(int temp)
{
  g_pTileSpace->SetTemp(temp);
  if (g_pTASApp && g_pTASApp->m_pEditorFrame && g_pTASApp->m_pEditorFrame->m_pTTWin) {
    g_pTASApp->m_pEditorFrame->m_pTTWin->SetTemps();
  }
  if (m_pSimFrame && m_pSimFrame->m_panel) {
    m_pSimFrame->m_panel->UpdateSpace(true);
  }
}

void TASApp::Set2HAMStageProgress(int percent)
{
  if (m_pSimFrame) {
    m_pSimFrame->Post2HAMProgressEvent(percent);
  }
}

void TASApp::SetSimOutputText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pSimFrame) {
    wxCommandEvent evt(tasEVT_POST_SETOUTPUTTEXT, ID_TAS_SIMULATOR_FRAME);

    wxDateTime now = wxDateTime::UNow();
    evt.SetString(now.Format(_T("[%c.%l] "), wxDateTime::Local) + sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pSimFrame, evt);
  }
}

void TASApp::SetSimDebugText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pSimFrame) {
    wxCommandEvent evt(tasEVT_POST_SETDEBUGTEXT, ID_TAS_SIMULATOR_FRAME);

    wxDateTime now = wxDateTime::UNow();
    evt.SetString(now.Format(_T("[%c.%l] "), wxDateTime::Local) + sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pSimFrame, evt);
  }
}

void TASApp::SetEditorOutputText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pEditorFrame) {
    wxCommandEvent evt(tasEVT_POST_SETOUTPUTTEXT, ID_TAS_EDITOR_FRAME);

    wxDateTime now = wxDateTime::UNow();
    evt.SetString(now.Format(_T("[%c.%l] "), wxDateTime::Local) + sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pEditorFrame, evt);
  }
}

void TASApp::SetEditorDebugText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pEditorFrame) {
    wxCommandEvent evt(tasEVT_POST_SETDEBUGTEXT, ID_TAS_EDITOR_FRAME);

    wxDateTime now = wxDateTime::UNow();
    evt.SetString(now.Format(_T("[%c.%l] "), wxDateTime::Local) + sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pEditorFrame, evt);
  }
}


void displayHelp() {
  //MJP: TODO - to stdout or a dialog?
}

void TASApp::ProcessCommandLineFiles(int argc, wxChar* argv[])
{
  bool bHadTdp = false;
  int count = 0;
  while (count < argc) {
    wxString sArg = argv[count];
    if (sArg.EndsWith(_T(".tdp"))) {
      g_pContext->m_fnTilesetFile = _T("");
      g_pContext->m_fnTilespaceFile = sArg;
      bHadTdp = true;
    } else if (sArg.EndsWith(_T(".tds"))) {
      g_pContext->m_fnTilespaceFile = _T("");
      g_pContext->m_fnTilesetFile = sArg;
    }
    count++;
  }
  if (bHadTdp && (g_pContext->m_fnTilesetFile == _T(""))) {
    wxString sPath = g_pContext->m_fnTilespaceFile.GetFullPath();
    g_pContext->m_fnTilesetFile = GetTdsFileNameFromTdpFile(sPath);
  }
}

void TASApp::SaveUserSettings()
{
  if (g_pContext->m_bDoSave) {
    g_pContext->SaveCurrentSettings();
  }
}

void TASApp::Srand(unsigned long seed, bool bRandomSeed) {
	if(bRandomSeed == true) {
		seed = time(0);
	}
	engine.seed(seed);
	m_randSeed = seed;
}

// Returns a random number between [0,n)
int TASApp::RandInRange(int n) {
	return static_cast<int>(RandomDouble() * n);
}

// Returns a random double between [0,1)
double TASApp::RandomDouble() {
	return distribution(engine);
}

int TASApp::OnExit()
{
  if (g_pTileSpaceBackup) {
    delete g_pTileSpaceBackup;
  }
  if (g_p2HAMTileSpace) {
    delete g_p2HAMTileSpace;
  }
  if (g_pkTAMTileSpace) {
	delete g_pkTAMTileSpace;
  }
  if (g_pEditorTileset) {
    delete g_pEditorTileset;
  }

/*
  if (g_printData) {
    delete g_printData;
  }
  if (g_pageSetupData) {
    delete g_pageSetupData;
  }
*/

  //  Clean up crash detection entry
  g_pContext->DeleteEntry(TASCrashCheck);

  if (g_pContext) {
    delete g_pContext;
  }

  return 0;
}

