/////////////////////////////////////////////////////////////////////////////
// Name:        TASEditor.h
// Purpose:     Declares frame and other classes used by editor
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _TAS_EDITOR_H
#define _TAS_EDITOR_H

#include <stdarg.h>
#include "TAS.h"
#include "TASFrmBase.h"
#include "TASPanel.h"
#include "TASTilePanel.h"
#include <wx/splitter.h>
#include <wx/frame.h>
#include <wx/scrolwin.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/fdrepdlg.h>
#include <wx/aui/aui.h>
#include <wx/filename.h>


#define TAS_IMAGE_TYPE_SVG  51

class TASSimulatorFrame;
class TASEditorFrame;
class TASTileEditor;
class TASNotebookWin;

class TASTiletypePanel
{
public:
  TASTiletypePanel();
  ~TASTiletypePanel();

  wxFileName        m_fn;
  TASTileEditor*    m_pEditor;
};

class TASTiletypePanels : public std::vector<TASTiletypePanel*>
{
};

#define TAS_IMAGE_DLG_TILETYPES 0
#define TAS_IMAGE_DLG_TILESPACE 1

class TASTileImageDialog : public wxDialog
{
public:
  TASTileImageDialog(wxWindow* parent, int type);

  void OnOk(wxCommandEvent& evt);
  void OnDirectoryButton(wxCommandEvent& evt);
  void OnText(wxCommandEvent& evt);

  wxTextCtrl* m_pDirText;
  wxRadioBox* m_pRadioMode;

  wxString      m_sDir;
  wxString      m_sExt;
  int           m_iDlgType;
  wxBitmapType  m_iImgType;

private:
  // any class wishing to process wxWidgets events must use this macro
  DECLARE_EVENT_TABLE()
};

/*
class TASEditorTextCtrl : public wxStyledTextCtrl
{
  friend class TASEditorTextPrint;
  
public:
  TASEditorTextCtrl(wxWindow* parent,
                       TASEditorFrame* parentFrame,
                       wxWindowID id,
                       const wxString& value = wxEmptyString,
                       const wxPoint& pos = wxDefaultPosition,
                       const wxSize& size = wxDefaultSize,
                       long style = 0,
                       const wxValidator& validator = wxDefaultValidator,
                       const wxString& name = wxTextCtrlNameStr);

  void OnTextMod(wxStyledTextEvent& event);
  void OnChar(wxKeyEvent& event);
  void OnEditClear(wxCommandEvent& event);
  void OnEditCut(wxCommandEvent& event);
  void OnEditCopy(wxCommandEvent& event);
  void OnEditPaste(wxCommandEvent& event);
  void OnEditIndentInc(wxCommandEvent& event);
  void OnEditIndentRed(wxCommandEvent& event);
  void OnEditSelectAll(wxCommandEvent& event);
  void OnEditSelectLine(wxCommandEvent& event);
  void OnEditRedo(wxCommandEvent& event);
  void OnEditUndo(wxCommandEvent& event);
  void OnBraceMatch(wxCommandEvent& event);
  void OnMarginClick(wxStyledTextEvent &event);
  void OnCommentSelection(wxCommandEvent& event);
  void OnUncommentSelection(wxCommandEvent& event);
  void OnKey(wxKeyEvent& event);

  bool InitializePrefs(const wxString &name);
  void SetSavePointEx();

private:
  TASEditorFrame*  m_parentFrame;

  // language properties
  LanguageInfo const* m_language;
  
  // margin variables
  int m_LineNrID;
  int m_LineNrMargin;
  int m_FoldingID;
  int m_FoldingMargin;
  int m_DividerID;

  DECLARE_CLASS(TASEditorTextCtrl)
    // any class wishing to process wxWindows events must use this macro
  DECLARE_EVENT_TABLE() 
};
*/

class TASEditorFrame : public TASFrameBase
{
public:
  TASEditorFrame(wxWindowID id, TASApp* parent, wxPoint& ptLoc, wxSize& szSize);
  ~TASEditorFrame();
  
  // Event handlers
  void OnNewTileset(wxCommandEvent& event);
  void OnLoadTileset(wxCommandEvent& event);
  void OnReloadTileset(wxCommandEvent& event);
  void OnSaveTileset(wxCommandEvent& event);
  void OnSaveTilesetAs(wxCommandEvent& event);

  void OnSaveTilesetImages(wxCommandEvent& event);

  void OnPrintSetup(wxCommandEvent& event);
  void OnPrintPreview(wxCommandEvent& event);
  void OnPrint(wxCommandEvent& event);
  void OnEdit(wxCommandEvent &event);
  void OnTilesetToSimulator(wxCommandEvent& event);
  void OnProcessSeed(wxCommandEvent& event);
  void OnClose(wxCloseEvent& event);
/*
  void OnFind(wxCommandEvent& event);
  void OnFindNext(wxCommandEvent& event);
  void OnFindPrev(wxCommandEvent& event);
  void OnReplace(wxCommandEvent& event);
  void OnReplaceNext(wxCommandEvent& event);
  void OnFindDialog(wxFindDialogEvent& event);
  void OnBookmarkAdd(wxCommandEvent& event);
  void OnBookmarkRem(wxCommandEvent& event);
  void OnNextBookmark(wxCommandEvent& event);
  void OnPrevBookmark(wxCommandEvent& event);
*/
  void OnToggleShowMessages(wxCommandEvent& event);
  void OnToggleShowTTEditor(wxCommandEvent& event);
  void OnToggleShowSymbols(wxCommandEvent& event);
  void OnToggleShowTileNames(wxCommandEvent& event);
  void OnToggleShowTileGlues(wxCommandEvent& event);
  void OnTileEditorEditMenu(wxCommandEvent& event);

  void SetDimMode();
  void SetTitle();
  void FindNext();
  wxRect DeterminePrintSize();
//  void SetEditorCursorPosition(int pos);
  void SetFileMenu();
  void SetEditMenu();
  void SetTilesetMenu();
//  void SetTabTitle(int tabNum = -1);
  int GetCurrentTab() { return m_pEditorFrameNB->GetSelection(); };

//  bool HasUnsavedWork();

  void ProcessUpdatedTileset();

//  TASEditorTextCtrl* GetCurrentTextCtrl();

  wxFindReplaceData*      m_findData;
  wxAuiNotebook*          m_pEditorFrameNB;
  wxMenu*                 m_pViewMenu;
  TASTiletypePanels       m_tileEditorPanels;
  TASSimulatorFrame*      m_pParentFrame;
  TASTileTypeWin*         m_pTTWin;

private:
  wxMenuBar*              m_pMenuBar;
  wxMenu*                 m_pFileMenu;
  wxMenu*                 m_pEditMenu;
  wxMenu*                 m_pTilesetMenu;
//  TASEditorTextCtrl*      m_ctrlEditorText;
//  TASEditorTextCtrl*      m_ctrlAnalyzerText;

  wxFindReplaceDialog*    m_dlgFind;
  wxFindReplaceDialog*    m_dlgReplace;

  DECLARE_CLASS(TASEditorFrame)
    // any class wishing to process wxWindows events must use this macro
  DECLARE_EVENT_TABLE()
};

#if wxUSE_PRINTING_ARCHITECTURE

/*
//----------------------------------------------------------------------------
//! TASEditorTextPrint
class TASEditorTextPrint: public wxPrintout
{  
public:
  
  //! constructor
  TASEditorTextPrint(TASEditorTextCtrl* edit, wxChar* title = _T(""));
  
  //! event handlers
  bool OnPrintPage(int page);
  bool OnBeginDocument(int startPage, int endPage);
  
  //! print functions
  bool HasPage(int page);
  void GetPageInfo(int* minPage, int* maxPage, int* selPageFrom, int* selPageTo);
  
private:
  TASEditorTextCtrl*   m_edit;
  int                     m_printed;
  wxRect                  m_pageRect;
  wxRect                  m_printRect;
  
  bool PrintScaling(wxDC* dc);
};
*/

#endif // wxUSE_PRINTING_ARCHITECTURE

class TASTextEntryOption {
public:
  TASTextEntryOption(const wxChar* m, const wxChar* v, int w) { m_message = m, m_value = v, m_width = w;}
  TASTextEntryOption() { m_width = 0; }
  wxString m_message;
  wxString m_value;
  int m_width;
};
WX_DECLARE_OBJARRAY(TASTextEntryOption, TASTextEntryOptionArray);

class TASTextEntryDlgData {
public:
  TASTextEntryDlgData(wxString& value, wxTextCtrl* pTextCtrl) { m_value = value; m_pTextCtrl = pTextCtrl; }
  wxString    m_value;
  wxTextCtrl* m_pTextCtrl;
};
WX_DECLARE_OBJARRAY(TASTextEntryDlgData, TASTextEntryDlgDataArray);

class TASTextEntryDialog : public wxDialog
{
public:
  TASTextEntryDialog(wxWindow *parent, const wxString& caption, const wxString& mainText,
                       TASTextEntryOptionArray& tteoa, long style = wxOK | wxCANCEL | wxCENTRE,
                       const wxPoint& pos = wxDefaultPosition);
  
  wxString GetValue(int valueNum) const;
  void OnOK(wxCommandEvent& event);
  
protected:
  TASTextEntryDlgDataArray m_dataArray;
  long m_dialogStyle;
  
private:
  DECLARE_EVENT_TABLE()
  DECLARE_DYNAMIC_CLASS(TASTextEntryDialog)
};

enum
{
  // menu items
  Menu_File_NewTileset = 300,
  Menu_File_LoadTileset,
  Menu_File_ReloadTileset,
  Menu_File_SaveTileset,
  Menu_File_SaveTileSeed,
  Menu_File_SaveTilesetAs,
  Menu_File_TilesetImages,
  Menu_View_ToggleShowMessages,
  Menu_View_ToggleShowTTEditor,
  Menu_View_ToggleShowSymbols,
  Menu_View_ToggleShowTileNames,
  Menu_View_ToggleShowTileGlues,

  Menu_Edit_ID_FINDNEXT,
  Menu_Edit_ID_FINDPREV,
  Menu_Edit_ID_REPLACE,
  Menu_Edit_ID_REPLACENEXT,
  Menu_Edit_ID_GOTO,
  Menu_Edit_ID_INDENTINC,
  Menu_Edit_ID_INDENTRED,
  Menu_Edit_ID_SELECTLINE,
  Menu_Tileset_IntoSimulator,
  Menu_Edit_ID_BRACEMATCH,
  Menu_Edit_ID_COMMENT,
  Menu_Edit_ID_UNCOMMENT,
  Menu_Edit_ID_BOOKMARKADD,
  Menu_Edit_ID_BOOKMARKREM,
  Menu_Edit_ID_NEXTBOOKMARK,
  Menu_Edit_ID_PREVBOOKMARK,

  Menu_Help,

  Dialog_Button_ID_IMAGEDIR,
  Dialog_Radio_ID_IMAGETYPE,
  Dialog_Text_ID_IMAGEDIR
};


#endif /* _TAS_EDITOR_H */



