// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

// For wxCURSOR_BLANK below, but isn't used a.t.m.
//#ifdef WIN32
//#include "wx/msw/wx.rc"
//#endif

#include "TASOSG.h"

#include <iostream>

#include <osg/Geode>
#include <osg/PositionAttitudeTransform>
#include <osg/Material>
#include <osg/Texture2D>
#include <osgViewer/Viewer>
#include <osg/Math>
#include <osg/Viewport>
#include <osg/PolygonMode>

#include <osgViewer/ViewerEventHandlers>
#include <osgGA/TrackballManipulator>
#include <osgDB/ReadFile>
#include <wx/image.h>
#include <wx/menu.h>
#include <wx/fileconf.h>
#include <wx/wfstream.h>
#include <wx/log.h>
#include <wx/tokenzr.h>
#include <wx/textfile.h>


osg::Vec3 TAS3DCubeFacePositions[] = { osg::Vec3(0.0f, 0.5, 0.0f), osg::Vec3(0.5f, 0.0f, 0.0f), osg::Vec3(0.0f, -0.5, 0.0f),
                                       osg::Vec3(-0.5f, 0.0, 0.0f), osg::Vec3(0, 0, 0.5f), osg::Vec3(0, 0, -0.5f) };

osg::Quat TAS3DCubeTextRotations[] = { osg::Quat(osg::DegreesToRadians(90.0), osg::Vec3d(1.0, 0.0, 0.0), osg::DegreesToRadians(180.0), osg::Vec3d(0.0, 0.0, 1.0), 0, osg::Vec3d(0,0,0)),
                                       osg::Quat(osg::DegreesToRadians(90.0), osg::Vec3d(0.0, 1.0, 0.0), osg::DegreesToRadians(90.0), osg::Vec3d(1.0, 0.0, 0.0), osg::DegreesToRadians(0.0), osg::Vec3d(0.0, 0.0, 0.0)),
                                       osg::Quat(osg::DegreesToRadians(90.0), osg::Vec3d(1.0, 0.0, 0.0)),
                                       osg::Quat(osg::DegreesToRadians(270.0), osg::Vec3d(0.0, 1.0, 0.0), osg::DegreesToRadians(90.0), osg::Vec3d(1.0, 0.0, 0.0), osg::DegreesToRadians(0.0), osg::Vec3d(0.0, 0.0, 0.0)),
                                       osg::Quat(),
                                       osg::Quat(osg::DegreesToRadians(180.0), osg::Vec3d(0.0, 1.0, 0.0), osg::DegreesToRadians(180.0), osg::Vec3d(0.0, .0, 1.0), osg::DegreesToRadians(0.0), osg::Vec3d(0.0, 0.0, 0.0)) };

//osg::Vec3 TAS3DCubeUnderlineOffsets[] = { osg::Vec3(0.0f, 0.5f, -0.1f), osg::Vec3(0.5f, 0.0f, -0.1f), osg::Vec3(0.0f, -0.5f, -0.1f), osg::Vec3(-0.5f, 0.0f, -0.1f), osg::Vec3(0.0f, -0.1f, 0.5f), osg::Vec3(0.0f, 0.1f, -0.5f) };
osg::Vec3 TAS3DCubeUnderlineOffsetsMult[] = { osg::Vec3(0,0,-0.1f), osg::Vec3(0,0,-0.1f), osg::Vec3(0,0,-0.1f), osg::Vec3(0,0,-0.1f), osg::Vec3(0,-0.1f,0), osg::Vec3(0,0.1f,0) };

int GetDirNumFromString(wxString str)
{
  if (str.CmpNoCase(_T("NORTH")) == 0) {
    return LNPS_NORTH;
  } else if (str.CmpNoCase(_T("SOUTH")) == 0) {
    return LNPS_SOUTH;
  } else if (str.CmpNoCase(_T("EAST")) == 0) {
    return LNPS_EAST;
  } else if (str.CmpNoCase(_T("WEST")) == 0) {
    return LNPS_WEST;
  } else if (str.CmpNoCase(_T("UP")) == 0) {
    return LNPS_UP;
  } else if (str.CmpNoCase(_T("DOWN")) == 0) {
    return LNPS_DOWN;
  } else {
    return -1;
  }
}

int GetIndexInPlane(int dir, int plane)
{ 
  for (int i=0; i<4; i++) {
    if (arrPlanes[plane][i] == dir) {
      return i;
    }
  }
  return -1;
}

int GetOrthogonalDir(int dir1, int dir2)
{ 
  if (abs(dir1 - dir2) == 3) {
    return (dir1 + 1) % LNPS_MAX_ARMS;    //  There are multiple valid choices here, just return the "first"
  } else {
    return (3 - ((dir1 % 3) + (dir2 % 3)));
  }
}

int GetAxis(int dir)
{
  return ((dir == LNPS_NORTH) || (dir == LNPS_SOUTH)) ? LNP_Y_AXIS :
          (((dir == LNPS_EAST) || (dir == LNPS_WEST)) ? LNP_X_AXIS : LNP_Z_AXIS );
}

bool IsOrthogonalAxis(int dir1, int dir2, int axis)
{
  if (abs(dir1 - dir2) == 3) {
    return (axis != GetAxis(dir1));
  } else {
    return (axis == GetAxis(GetOrthogonalDir(dir1, dir2)));
  }
}

////////////////////////////////////////////////////////////////////////////
class AxisCameraUpdateCallback:public osg::NodeCallback
{
public:
    virtual void operator()(osg::Node* node, osg::NodeVisitor* nv)
    {
        if(nv->getVisitorType() == osg::NodeVisitor::UPDATE_VISITOR)
        {
            osg::Camera* camera = dynamic_cast<osg::Camera*>(node);
            if (camera)
            {
               osg::View* view = camera->getView();
               if (view && view->getNumSlaves() > 0)
               {
                   osg::View::Slave* slave = &view->getSlave(0);
                   if(slave->_camera.get() == camera)
                   {
                       osg::Camera* masterCam = view->getCamera();
                       osg::Vec3 eye, center, up;
                       masterCam->getViewMatrixAsLookAt(eye, center, up, 30);
                       osg::Matrixd matrix;
                       matrix.makeLookAt(eye-center, osg::Vec3(0, 0, 0), up); // always look at (0, 0, 0)
                       camera->setViewMatrix(matrix);
                   }
               }
            }
        }
        traverse(node,nv);
    }
};

osg::Camera* createHUD()
{
   osg::Camera* camera = new osg::Camera;
   camera->setProjectionMatrix(osg::Matrix::ortho2D(-1.5, 17.75, -1.5, 13.86));
   camera->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
   camera->setClearMask(GL_DEPTH_BUFFER_BIT);
   camera->setRenderOrder(osg::Camera::POST_RENDER);
   camera->setUpdateCallback(new AxisCameraUpdateCallback);

   osg::Node* axes = osgDB::readNodeFile("TAS-axes.osg");
   camera->addChild(axes);

   return camera;
}
////////////////////////////////////////////////////////////////////////////

//   Force the Node and its children to be wireframe, overrides parents state
void forcedWireFrameModeOn( osg::Node *srcNode ){
  // Quick sanity check , we need a node
  //
  if( srcNode == NULL )
    return;
 
  // --------------------------------------------------------------------------------
  //   Grab the state set of the node, this  will a StateSet if one does not exist
  // --------------------------------------------------------------------------------
  osg::StateSet *state = srcNode->getOrCreateStateSet();
 
  // ----------------------------------------------------------------------------------------------
  //   We need to retireve the Poylgon mode of the  state set, and create one if does not have one
  // ----------------------------------------------------------------------------------------------
  osg::PolygonMode* polyModeObj = dynamic_cast< osg::PolygonMode* >( state->getAttribute( osg::StateAttribute::POLYGONMODE ));
 
  if ( !polyModeObj ) {
    polyModeObj = new osg::PolygonMode;
    state->setAttribute( polyModeObj );    
  }
 
  // --------------------------------------------   
  //  Now we can set the state to WIREFRAME
  // --------------------------------------------   
  polyModeObj->setMode(  osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE );

}  // forceWireFrameModeOn

//MJP: turning this off doesn't seem to help performance much and makes visualization much worse
int showWireframe = 1;
int staticcount = 0;

osg::Group* LNP::CreateOSGObject()
{
  osg::Group* group = new osg::Group();

  osg::StateSet* stateset = new osg::StateSet();
  group->setStateSet(stateset);

  osg::TessellationHints* hints = new osg::TessellationHints;
  hints->setDetailRatio(1.0f);

  //  Create cube
  float dim = 0.99f;
  if (showWireframe != 1) {
    dim = 1.0f;
  }
  osg::Box* pBox = new osg::Box(osg::Vec3(0.0f,0.0f,0.0f), dim);
  osg::ShapeDrawable* pBoxDrawable = new osg::ShapeDrawable(pBox, hints);
  pBoxDrawable->setColor(osg::Vec4(255, 255, 255, 1.0));

  osg::Geode* boxGeode = new osg::Geode();
  std::string sName;
  sName += staticcount++;
  boxGeode->setName(sName);
  //boxGeode->setUpdateCallback(
  boxGeode->setStateSet(stateset);
  boxGeode->addDrawable(pBoxDrawable);

  //MJP: use smart pointers!
  osg::Geode* boxGeode2 = NULL;
  if (showWireframe == 1) {
    //  Create wireframe
    osg::StateSet* stateset2 = new osg::StateSet();
    group->setStateSet(stateset);

    osg::Box* pBox2 = new osg::Box(osg::Vec3(0.0f,0.0f,0.0f), 1.0f);
    osg::ShapeDrawable* pBoxDrawable2 = new osg::ShapeDrawable(pBox2, hints);
    pBoxDrawable2->setColor(osg::Vec4(0, 0, 0, 0.1));

    boxGeode2 = new osg::Geode();
    boxGeode2->setStateSet(stateset2);
    forcedWireFrameModeOn(boxGeode2);
    boxGeode2->addDrawable(pBoxDrawable2);
  }


  std::string tempText[] = { "North", "East", "South", "West", "Up", "Down" };

  for (int i=TAS_NORTH; i<=TAS_DOWN; i++) {

    std::string sLabel = tempText[i];
    osg::Geode* textGeode = new osg::Geode();
    osgText::Text* text = new osgText::Text();
    //text->setFont("tahoma.ttf");
    text->setCharacterSize(0.5f);
    text->setPosition(TAS3DCubeFacePositions[i]);
    text->setRotation(TAS3DCubeTextRotations[i]);
    text->setAlignment(osgText::Text::CENTER_BASE_LINE);
    text->setColor(osg::Vec4(0.0f, 0.0f, 0.0f, 1.0f));
    text->setText(sLabel);
    textGeode->addDrawable(text);

    for (int j=1; j<=2; j++) {
      osgText::Text* text2 = new osgText::Text(*text);
      osg::Vec3 v(TAS3DCubeFacePositions[i]);
      osg::Vec3 m(TAS3DCubeUnderlineOffsetsMult[i]*j*0.7f);
      v += m;
      text2->setPosition(v);
      std::string sULine;
      sULine.resize(sLabel.length(), '-');
      text2->setText(sULine);
      textGeode->addDrawable(text2);
    }

    group->addChild(textGeode);
  }


  //  Add cube
  osg::PositionAttitudeTransform* boxPAT = new osg::PositionAttitudeTransform();
  boxPAT->setPosition(osg::Vec3(0.0,0,0));
  boxPAT->addChild(boxGeode);
  if (showWireframe == 1) {
    boxPAT->addChild(boxGeode2);
  }
  group->addChild(boxPAT);

  return group;
}


/*class PointHolder {
public:
  PointHolder() { m_xDir = m_yDir = m_zDir = 1; m_bValid = true; };
  PointHolder(int xDir, int yDir, int zDir);

  std::vector<std::vector<std::vector<TASCoords>>> m_points;
  bool  m_bValid;
  int   m_xDir;
  int   m_yDir;
  int   m_zDir;
};

PointHolder::PointHolder(int xDir, int yDir, int zDir) {
  m_bValid = false;

  if ((xDir == 1) || (xDir == -1)) {
    m_xDir = xDir;
  } else {
    return;
  }
  if ((yDir == 1) || (yDir == -1)) {
    m_yDir = yDir;
  } else {
    return;
  }
  if ((zDir == 1) || (zDir == -1)) {
    m_zDir = zDir;
  } else {
    return;
  }

  m_bValid = true;
}*/

bool FractalViewerFrame::Grow()
{
  double d = m_stage;
  m_stage++;
  size_t mult = pow(m_scale, d);
  osg::Vec3 origin(0,0,0);
  size_t currSize = m_lnpAssembly.size();
  osg::ref_ptr<osg::Group> pNewGroup = new osg::Group();
  for (size_t j=0; j<m_generator.size(); j++) {
    if (m_generator[j] != origin) {
      AddCopyAtOffset(pNewGroup, m_generator[j]*mult);
    }
    /*  MJP: more efficiently utilize OSG by making an additional copy of the current group for each point in the generator
    if (m_generator[j] == origin) {

    } else {
      for (size_t i=0; i<currSize; i++) {
        osg::Vec3 c = m_lnpAssembly[i].m_vecLoc + (m_generator[j]*mult);
        AddLNP(c);
      }
    }*/
  }
  m_pGroup->addChild(pNewGroup);

  double numPoints = pow(m_generator.size(), (double)m_stage);

  wxString sMsg;
  sMsg = sMsg.Format(wxT("Fractal stage: %d, with %.0f points"), m_stage, numPoints);
  SetStatusText(sMsg, 0);
  return true;
}

LNPBaseFrame::LNPBaseFrame(TASApp* pApp, wxFrame *frame, const wxString& title, const wxPoint& pos,
    const wxSize& size, long style)
    : wxFrame(frame, wxID_ANY, title, pos, size, style)
{
  m_pApp   = pApp;
  m_pGroup = new osg::Group();

  m_pStatusBar = NULL;

  // create osg canvas
  //    - initialize
  int *attributes = new int[7];
  attributes[0] = int(WX_GL_DOUBLEBUFFER);
  attributes[1] = WX_GL_RGBA;
  attributes[2] = WX_GL_DEPTH_SIZE;
  attributes[3] = 8;
  attributes[4] = WX_GL_STENCIL_SIZE;
  attributes[5] = 8;
  attributes[6] = 0;

  OSGCanvas *canvas = new OSGCanvas(this, wxID_ANY, wxDefaultPosition,
      wxSize(size.GetWidth(), size.GetHeight()), wxSUNKEN_BORDER, wxT("FractalViewer"), attributes);

  GraphicsWindowWX* gw = new GraphicsWindowWX(canvas);

  canvas->SetGraphicsWindow(gw);

  osgViewer::Viewer *viewer = new osgViewer::Viewer;
  viewer->getCamera()->setGraphicsContext(gw);
  viewer->getCamera()->setViewport(0,0,size.GetWidth(),size.GetHeight());
  viewer->addEventHandler(new osgViewer::StatsHandler);
  viewer->setThreadingModel(osgViewer::Viewer::SingleThreaded);

  // add the handler for doing the picking
  osg::ref_ptr<osgText::Text> updateText = new osgText::Text;
  viewer->addEventHandler(new PickHandler(updateText.get(), this));

  ////////////////////////////////////////////////////////////////////////////
  //osg::Camera* camera = createHUD();
  //viewer->addSlave(camera, false);
  //m_pGroup->addChild(camera);
  ////////////////////////////////////////////////////////////////////////////

  viewer->setSceneData(m_pGroup);

  viewer->setCameraManipulator(new osgGA::TrackballManipulator);
  SetViewer(viewer);
}

void LNPBaseFrame::SetViewer(osgViewer::Viewer *viewer)
{
  _viewer = viewer;
}

void LNPBaseFrame::OnIdle(wxIdleEvent &event)
{
  _viewer->frame();

  event.RequestMore();
}

// Event tables
BEGIN_EVENT_TABLE(FractalViewerFrame, wxFrame)
    EVT_IDLE(FractalViewerFrame::OnIdle)
END_EVENT_TABLE()

/* My frame constructor */
FractalViewerFrame::FractalViewerFrame(TASApp* pApp, wxFrame *frame, const wxString& title, const wxPoint& pos, const wxSize& size, long style)
                : LNPBaseFrame(pApp, frame, title, pos, size, style)
{
  wxMenu *menuFile = new wxMenu(wxMENU_TEAROFF);

  menuFile->Append(wxID_NEW, wxEmptyString, _("Start a new fractal"));
  menuFile->AppendSeparator();
  menuFile->Append(wxID_OPEN, wxEmptyString, _("Open a fractal definition"));
  menuFile->AppendSeparator();
  menuFile->Append(wxID_SAVE, wxEmptyString, _("Save fractal definition"));
  menuFile->Append(wxID_SAVEAS, wxEmptyString, _("Save fractal definition as.."));

  menuFile->AppendSeparator();
  menuFile->Append(wxID_EXIT);

  wxMenuBar *menuBar = new wxMenuBar();
  menuBar->Append(menuFile, _("&File"));
  SetMenuBar(menuBar);

  m_stage      = 0;
  m_pStatusBar = CreateStatusBar(1);
  SetStatusText(_("No fractal loaded"));

  m_pLogWin = new wxLogWindow(this, wxT("Messages"));
  wxFrame* pFrame = m_pLogWin->GetFrame();
  pFrame->Move(200, 750);
}

BoxLoc::BoxLoc(LNP* pLNP, osg::Vec3 vecLoc)
{
  m_lnp       = *pLNP;
  m_vecLoc    = vecLoc;

  m_pat = new osg::PositionAttitudeTransform();
  m_pat->setPosition(vecLoc);

  osg::Group* pGroup = m_lnp.CreateOSGObject();
  m_pat->addChild(pGroup);
}

bool FractalViewerFrame::AddLNP(osg::Vec3& coords)
{
  BoxLoc loc(&m_lnp, coords);
  m_lnpAssembly.push_back(loc);
  m_pGroup->addChild(loc.m_pat);

  return true;
}

bool FractalViewerFrame::AddGroup(std::vector<osg::Vec3>* pVectCoords)
{
  bool bRet = true;
  for (size_t i=0; i<pVectCoords->size(); i++) {
    bRet = (bRet && AddLNP(pVectCoords->at(i)));
  }
  return bRet;
}

bool FractalViewerFrame::AddCopyAtOffset(osg::ref_ptr<osg::Group> pParentGroup, osg::Vec3 coords)
{
  osg::ref_ptr<osg::Group> newGroup = new osg::Group(*m_pGroup);//, osg::CopyOp::DEEP_COPY_OBJECTS);
  osg::PositionAttitudeTransform* boxPAT = new osg::PositionAttitudeTransform();
  boxPAT->setPosition(coords);
  boxPAT->addChild(newGroup);
  pParentGroup->addChild(boxPAT);

  return true;
}
 

BEGIN_EVENT_TABLE(OSGCanvas, wxGLCanvas)
    EVT_SIZE                (OSGCanvas::OnSize)
    EVT_PAINT               (OSGCanvas::OnPaint)
    EVT_ERASE_BACKGROUND    (OSGCanvas::OnEraseBackground)

    EVT_CHAR                (OSGCanvas::OnChar)
    EVT_KEY_UP              (OSGCanvas::OnKeyUp)

    EVT_ENTER_WINDOW        (OSGCanvas::OnMouseEnter)
    EVT_LEFT_DOWN           (OSGCanvas::OnMouseDown)
    EVT_MIDDLE_DOWN         (OSGCanvas::OnMouseDown)
    EVT_RIGHT_DOWN          (OSGCanvas::OnMouseDown)
    EVT_LEFT_UP             (OSGCanvas::OnMouseUp)
    EVT_MIDDLE_UP           (OSGCanvas::OnMouseUp)
    EVT_RIGHT_UP            (OSGCanvas::OnMouseUp)
    EVT_MOTION              (OSGCanvas::OnMouseMotion)
END_EVENT_TABLE()

OSGCanvas::OSGCanvas(wxWindow *parent, wxWindowID id,
    const wxPoint& pos, const wxSize& size, long style, const wxString& name, int *attributes)
    : wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name, attributes)
{
    // default cursor to standard
    _oldCursor = *wxSTANDARD_CURSOR;
}

OSGCanvas::~OSGCanvas()
{
}

void OSGCanvas::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
    /* must always be here */
    wxPaintDC dc(this);
}

void OSGCanvas::OnSize(wxSizeEvent& event)
{
    // this is also necessary to update the context on some platforms
    wxGLCanvas::OnSize(event);

    // set GL viewport (not called by wxGLCanvas::OnSize on all platforms...)
    int width, height;
    GetClientSize(&width, &height);

    if (_graphics_window.valid())
    {
        // update the window dimensions, in case the window has been resized.
        _graphics_window->getEventQueue()->windowResize(0, 0, width, height);
        _graphics_window->resized(0,0,width,height);
    }
}

void OSGCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
    /* Do nothing, to avoid flashing on MSW */
}

void OSGCanvas::OnChar(wxKeyEvent &event)
{
#if wxUSE_UNICODE
    int key = event.GetUnicodeKey();
#else
    int key = event.GetKeyCode();
#endif

    if (_graphics_window.valid())
        _graphics_window->getEventQueue()->keyPress(key);

    // If this key event is not processed here, we should call
    // event.Skip() to allow processing to continue.
}

void OSGCanvas::OnKeyUp(wxKeyEvent &event)
{
#if wxUSE_UNICODE
    int key = event.GetUnicodeKey();
#else
    int key = event.GetKeyCode();
#endif

    if (_graphics_window.valid())
        _graphics_window->getEventQueue()->keyRelease(key);

    if ((key == 'g') || (key == 'G')) {
      FractalViewerFrame* pFrame = (FractalViewerFrame*)GetParent();
      pFrame->Grow();
    }

    // If this key event is not processed here, we should call
    // event.Skip() to allow processing to continue.
    event.Skip();
}

void OSGCanvas::OnMouseEnter(wxMouseEvent &event)
{
    // Set focus to ourselves, so keyboard events get directed to us
    SetFocus();
}

void OSGCanvas::OnMouseDown(wxMouseEvent &event)
{
    if (_graphics_window.valid())
    {
        _graphics_window->getEventQueue()->mouseButtonPress(event.GetX(), event.GetY(),
            event.GetButton());
    }
}

void OSGCanvas::OnMouseUp(wxMouseEvent &event)
{
    if (_graphics_window.valid())
    {
        _graphics_window->getEventQueue()->mouseButtonRelease(event.GetX(), event.GetY(),
            event.GetButton());
    }
}

void OSGCanvas::OnMouseMotion(wxMouseEvent &event)
{
    if (_graphics_window.valid())
        _graphics_window->getEventQueue()->mouseMotion(event.GetX(), event.GetY());
}

void OSGCanvas::UseCursor(bool value)
{
    if (value)
    {
        // show the old cursor
        SetCursor(_oldCursor);
    }
    else
    {
        // remember the old cursor
        _oldCursor = GetCursor();

        // hide the cursor
        //    - can't find a way to do this neatly, so create a 1x1, transparent image
        wxImage image(1,1);
        image.SetMask(true);
        image.SetMaskColour(0, 0, 0);
        wxCursor cursor(image);
        SetCursor(cursor);

        // On wxGTK, only works as of version 2.7.0
        // (http://trac.wxwidgets.org/ticket/2946)
        // SetCursor( wxStockCursor( wxCURSOR_BLANK ) );
    }
}

GraphicsWindowWX::GraphicsWindowWX(OSGCanvas *canvas)
{
  _canvas = canvas;

  _traits = new GraphicsContext::Traits;

  wxPoint pos = _canvas->GetPosition();
  wxSize  size = _canvas->GetSize();

  _traits->x = pos.x;
  _traits->y = pos.y;
  _traits->width = size.x;
  _traits->height = size.y;

  init();
}

GraphicsWindowWX::~GraphicsWindowWX()
{
}

void GraphicsWindowWX::init()
{
    if (valid())
    {
        setState( new osg::State );
        getState()->setGraphicsContext(this);

        if (_traits.valid() && _traits->sharedContext)
        {
            getState()->setContextID( _traits->sharedContext->getState()->getContextID() );
            incrementContextIDUsageCount( getState()->getContextID() );
        }
        else
        {
            getState()->setContextID( osg::GraphicsContext::createNewContextID() );
        }
    }
}

void GraphicsWindowWX::grabFocus()
{
    // focus the canvas
    _canvas->SetFocus();
}

void GraphicsWindowWX::grabFocusIfPointerInWindow()
{
  // focus this window, if the pointer is in the window
  wxPoint pos = wxGetMousePosition();
  if (wxFindWindowAtPoint(pos) == _canvas) {
    _canvas->SetFocus();
  }
}

void GraphicsWindowWX::useCursor(bool cursorOn)
{
  _canvas->UseCursor(cursorOn);
}

bool GraphicsWindowWX::makeCurrentImplementation()
{
  _canvas->SetCurrent();
  return true;
}

void GraphicsWindowWX::swapBuffersImplementation()
{
  _canvas->SwapBuffers();
}

bool PickHandler::handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa)
{
    switch(ea.getEventType())
    {
//MJP TODO: provide text feedback of LNP addition
//MJP TODO: figure out how space-bar click re-centers and re-sizes image! KEY_Space
        /*case(osgGA::GUIEventAdapter::PUSH)://DOUBLECLICK
        {
          //ea.getButtonMask();
          //ea.getModKeyMask();
          //if ((ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_SHIFT) != 0) {
          if ((ea.getModKeyMask() != 0) || (ea.getButtonMask() != 0)) {
            osgViewer::View* view = dynamic_cast<osgViewer::View*>(&aa);
            if (view) pick(view, ea);
            return false;
          }
        }*/
        case(osgGA::GUIEventAdapter::KEYDOWN):
        {
          if (ea.getKey() == wxT('a')) {
            osgViewer::View* view = dynamic_cast<osgViewer::View*>(&aa);
            if (view) {
              if (pick(view, ea, LNP_PICK_ACTION_ADD)) {
                return false;
              }
            }
            return false;
          }
          else if (ea.getKey() == wxT('r')) {
            osgViewer::View* view = dynamic_cast<osgViewer::View*>(&aa);
            if (view) {
              if (pick(view, ea, LNP_PICK_ACTION_ROTATE)) {
                return false;
              }
            }
            return false;
          }
          else if (ea.getKey() == wxT('c')) {        

          }
          return false;
        }
        case(osgGA::GUIEventAdapter::MOVE):
        {
            osgViewer::View* view = dynamic_cast<osgViewer::View*>(&aa);
            if (view) {
              if (pick(view, ea, LNP_PICK_ACTION_IDENTIFY)) {
                return false;
              }
            }
        }
        default:
            return false;
    }
}

bool PickHandler::pick(osgViewer::View* view, const osgGA::GUIEventAdapter& ea, int action)
{
    osgUtil::LineSegmentIntersector::Intersections intersections;

//    std::string gdlist="";
    float x = ea.getX();
    float y = ea.getY();

//MJP: shortcut to get out of here for now!!!
return true;

    if (view->computeIntersections(x,y,intersections))
    {
        for(osgUtil::LineSegmentIntersector::Intersections::iterator hitr = intersections.begin();
            hitr != intersections.end();
            ++hitr)
        {
            std::ostringstream os;
            /*if (!hitr->nodePath.empty()) {
              osg::Node* pHitNode = hitr->nodePath.back();
              std::string s = pHitNode->getName();
            }*/
            if (!hitr->nodePath.empty() && !(hitr->nodePath.back()->getName().empty()))
            {
                // the geodes are identified by name.
//                os<<"Object \""<<hitr->nodePath.back()->getName()<<"\""<<std::endl;
                //wxString sMsg = wxString::Format(wxT("Hit object '%s'"), hitr->nodePath.back()->getName().c_str());
                //wxMessageBox(sMsg, "Hit object");
                osg::Node* pHitNode = hitr->nodePath.back();

                wxString sArm;
                wxString sName(pHitNode->getName());
                if (sName.Contains(wxT("Arm:"))) {
                  sArm = sName.Before(wxT(' '));
                }
                while (pHitNode && !sName.StartsWith(wxT("Base:"))) {
                  pHitNode = pHitNode->getParent(0);
                  sName = pHitNode->getName();
                  if (sName.Contains(wxT("Arm:"))) {
                    sArm = sName.Before(wxT(' '));
                    if (action == LNP_PICK_ACTION_ROTATE) {
                      wxLogMessage(wxT("Attempt to rotate around arm: %s"), sArm.c_str());
                      /*if (m_pFrame && m_pFrame->m_pApp && m_pFrame->m_pApp->m_pFractalViewerFrame) {
                        LNPAssembly* pAssembly = &m_pFrame->m_pApp->m_pFractalViewerFrame->m_lnpAssembly;
                        for (size_t i=0; i<pAssembly->size(); i++) {
                          wxLogMessage(wxT("(%f,%f,%f)"), pAssembly->at(i).m_vecLoc.x(), pAssembly->at(i).m_vecLoc.y(), pAssembly->at(i).m_vecLoc.z());
//MJP TODO:  This is the spot to figure out which arm we are attempting to rotate around and whether or not any other arms should prevent that
                        }
					  }*/
                    }
                  }
                }
                if (sName.StartsWith(wxT("Base:"))) {
                  sName = sName.AfterFirst(_T(':'));
                  /*if (m_pFrame && m_pFrame->m_pApp && m_pFrame->m_pApp->m_pFractalViewerFrame) {
                    if (action == LNP_PICK_ACTION_ADD) {
                      m_pFrame->m_pApp->m_pFractalViewerFrame->AddLNP(sName);
                      wxString sText = wxString::Format(wxT("Added '%s'"), sName.c_str());
                      BoxLoc* pLoc = &m_pFrame->m_pApp->m_pFractalViewerFrame->m_lnpAssembly[m_pFrame->m_pApp->m_pFractalViewerFrame->m_lnpAssembly.size()-1];
                      for (int i=0; i<LNPS_MAX_ARMS; i++) {
                        if (pLoc->m_lnp.m_arrArms[i].m_bBound) {
                          sText += wxT(" by glue ") + pLoc->m_lnp.m_arrArms[i].m_glue.m_sLabel;
                          break;
                        }
                      }
                      m_pFrame->m_pApp->m_pFractalViewerFrame->SetStatusText(sText, 0);
                    } else if (action == LNP_PICK_ACTION_IDENTIFY) {

                    }
                    return true;
                  }*/
                }
                break;
            }
/*
            else if (hitr->drawable.valid())
            {
                os<<"Object \""<<hitr->drawable->className()<<"\""<<std::endl;
            }

            os<<"        local coords vertex("<< hitr->getLocalIntersectPoint()<<")"<<"  normal("<<hitr->getLocalIntersectNormal()<<")"<<std::endl;
            os<<"        world coords vertex("<< hitr->getWorldIntersectPoint()<<")"<<"  normal("<<hitr->getWorldIntersectNormal()<<")"<<std::endl;
            const osgUtil::LineSegmentIntersector::Intersection::IndexList& vil = hitr->indexList;
            for(unsigned int i=0;i<vil.size();++i)
            {
                os<<"        vertex indices ["<<i<<"] = "<<vil[i]<<std::endl;
            }
            
            gdlist += os.str();
*/
        }
    }
//    setLabel(gdlist);

    /*if (action == LNP_PICK_ACTION_IDENTIFY) {
      m_pFrame->SetStatusText(wxT(""), 1);
    }*/
    return false;
}
