/////////////////////////////////////////////////////////////////////////////
// Name:        TASSimFrm.h
// Purpose:     Declares the frame used for the simulator window
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef TASSimFrame_H
#define TASSimFrame_H

#include <wx/dcclient.h>
#include <wx/dynarray.h>
#include <wx/button.h>
#include <wx/checkbox.h>
#include <wx/defs.h>
#include <wx/frame.h>
#include <wx/sizer.h>
#include <wx/radiobox.h>
#include <wx/radiobut.h>
#include <wx/thread.h>
#include <wx/popupwin.h>
#include <wx/scrolwin.h>
#include <wx/choice.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/aui/aui.h>
#ifdef __WXMAC__
  #include <OpenGL/gl.h>
#else
  #include <GL/gl.h>
#endif /* __WXMAC__ */
#include "TASFrmBase.h"

class TASSimulatorFrame;

#include "TASEditor.h"
#include "TASPanel.h"


// forward declarations
class TASApp;
class TASEditorFrame;
class TASPanelBase;
class TAS3DPanel;
class TAS3DCellValueWin;
class TAS3DCASpaceWin;
class TAS3DAxesWin;
class TASNotebookWin;
class TAS2HAMConfigWin;


class TASIteratorThread : public wxThread
{
public:
  TASIteratorThread(wxThreadKind kind, TASSimulatorFrame* m_pSimFrame, wxMutex* pMutexIterate, wxCondition* pCondiitonIterate);
  ~TASIteratorThread();

  virtual void* Entry();
  virtual bool Iterate();
  virtual bool BackUp();

  bool GetShouldDie();
  void SetShouldDie();

  void PostUpdateEvent(bool bEntire, bool bAsync);
  void Post2HAMUpdateEvent();
  
  TASSimulatorFrame*  m_pSimFrame;
  wxMutex*            m_pMutexIterate;
  wxCondition*        m_pConditionIterate;
  bool                m_bShouldDie;
};


// Define a new frame type: this is going to be our main frame
class TASSimulatorFrame : public TASFrameBase
{
public:
  TASSimulatorFrame(wxWindowID id, TASApp* parentApp, const wxString& title, const wxPoint& pos, const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);
	virtual ~TASSimulatorFrame();

  // event handlers
  void OnSize(wxSizeEvent& event);
  void OnQuit(wxCommandEvent& event);
  void OnClose(wxCloseEvent& event);
  void OnAbout(wxCommandEvent& event);

  void OnNewTAS(wxCommandEvent& event);
  void OnLoadTAS(wxCommandEvent& event);
  void OnReloadTAS(wxCommandEvent& event);
  void OnSaveTAS(wxCommandEvent& event);
  void OnSaveTASAs(wxCommandEvent& event);
  void OnNewAssembly(wxCommandEvent& event);
  void OnLoadAssembly(wxCommandEvent& event);
  void OnReloadAssembly(wxCommandEvent& event);
  void OnSaveAssembly(wxCommandEvent& event);
  void OnSaveAssemblyAs(wxCommandEvent& event);

  void OnBatchMode(wxCommandEvent& event);
  void OnResetAssembly(wxCommandEvent& event);
  void OnLoadTilesetIntoSimulator(wxCommandEvent& event);
  void OnLoadTilesetFromSimulator(wxCommandEvent& event);

  void OnSaveAssemblyImage(wxCommandEvent& event);
  void OnSavePointSet(wxCommandEvent& event);

  void OnPrint(wxCommandEvent& event);
  void OnPrintPreview(wxCommandEvent& event);
  void OnPrintSetup(wxCommandEvent& event);
  void OnPageSetup(wxCommandEvent& event);
  void OnSetBackgroundColor(wxCommandEvent& event);
  void OnSetTileColor(wxCommandEvent& event);
  void OnSetTextColor(wxCommandEvent& event);

  void OnSetRefresh(wxCommandEvent& event);
  void OnSetViewType(wxCommandEvent& event);

  void OnControl(wxCommandEvent& event);
  void OnSimulatorReset(wxCommandEvent& event);
  void OnSetBreakpoint(wxCommandEvent& event);
  void OnSetAsServer(wxCommandEvent& event);
  void OnSetAsClient(wxCommandEvent& event);
  void OnClearDebugWindow(wxCommandEvent& event);
  void OnCopyCurrentTileType(wxCommandEvent& event);
  void OnFindCurrentTile(wxCommandEvent& event);
  void OnRefreshScreenNow(wxCommandEvent& event);
  void OnCheckLocalDeterminism(wxCommandEvent& event);
  void OnSelectDimension(wxCommandEvent& event);
  void OnSelectSimulationMode(wxCommandEvent& event);
  void OnSetTemperature(wxCommandEvent& event);
  void OnSetStageFile(wxCommandEvent& event);
  void OnUseTileConcentrations(wxCommandEvent& event);
  void OnSetTileAdditions(wxCommandEvent& event);
  void OnSetUse2HAMHack(wxCommandEvent& event);
  void OnCheckAysncDisplay(wxCommandEvent& event);
  void OnRandomSeed(wxCommandEvent& event);
  void OnClearUserSettings(wxCommandEvent& event);

  void OnToggleUseEditor(wxCommandEvent& event);
  void OnToggleSpaceConfig(wxCommandEvent& event);
  void OnToggle2HAMConfig(wxCommandEvent& event);
  void OnToggleCellValues(wxCommandEvent& event);
  void OnToggleShowMouseOver(wxCommandEvent& event);
  void OnToggleShowFrontier(wxCommandEvent& event);
  void OnToggleShowSymbols(wxCommandEvent& event);
  void OnToggleShowTileNames(wxCommandEvent& event);
  void OnToggleShowTileGlues(wxCommandEvent& event);
  void OnToggleShowAxes(wxCommandEvent& event);
  void OnToggleShowOverview(wxCommandEvent& event);
  void OnToggleShowMessages(wxCommandEvent& event);
  void OnZoomChange(wxCommandEvent& event);
  void OnZoomIn(wxCommandEvent& event);
  void OnZoomOut(wxCommandEvent& event);
  void OnModifyFrontier(wxCommandEvent& event);

  void OnToolEnter(wxCommandEvent& event);
  
  void OnToggleAutoRotate(wxCommandEvent& event);
  void OnOutputDebugMsgs(wxCommandEvent& event);
  void OnDebugOutput(wxCommandEvent& event);

  void OnPostUpdateSpaceIncr(wxCommandEvent& event);
  void OnPostUpdateSpaceIncrWithSig(wxCommandEvent& event);
  void OnPostUpdateSpaceFull(wxCommandEvent& event);
  void OnPostUpdateSpaceFullWithSig(wxCommandEvent& event);
  void OnPost2HAMUpdate(wxCommandEvent& event);
  void OnPost2HAMProgressUpdate(wxCommandEvent& event);
  
  void OnMouseWheel(wxMouseEvent& event);
//  void OnMouseRightDown(wxMouseEvent& event);
//  void OnMouseRightUp(wxMouseEvent& event);
//  void OnMouseMove(wxMouseEvent& event);
//  void OnMouseLeave(wxMouseEvent& event);

  void DoOnSize();
  void DoZoom(float fZoomFact);
  void DoRefresh();

  void BuildFileMenu();
  void BuildAppearanceMenu();
  void BuildOptionMenu();
  void BuildToolbar();
  void SetRefreshRateMenu(int rate);
  void SetTASDimMode(int mode, bool bForce = false);
  int  GetTASDimMode();
  void SetTASSimMode(int mode);
  int  GetTASSimMode();
  void SetTitle();
  void UpdateStatusBar();
  void Post2HAMProgressEvent(int percent);
  void SetZoomLevel(float fZoomLevel);
  bool GetPrintBounds(int* fRow, int* lRow, int* fCol, int* lCol);
  void AddAxesWin();
  void RemoveAxesWin();
  void RotateAxes(GLfloat* pQuat);
  void AddCellValueWin();
  void Add2HAMConfigWin();
  void Remove2HAMConfigWin();
  void AddSpaceConfigWin();
  void RemoveSpaceConfigWin();
  void AddOverviewWin();
  void RemoveOverviewWin();
  void RefreshOverviewWin();
  void SetWindowVisibleChecks();
  void SetUseTileConcCheck();

  void ProcessUpdatedTileset();
  void ProcessUpdatedTilespace();

  TASPanelBase*         m_panel;
  TAS3DCellValueWin*    m_pCellValueWin;
  TAS3DCASpaceWin*      m_pSpaceConfigWin;
  TAS2HAMConfigWin*     m_p2HAMConfigWin;
  TAS3DAxesWin*         m_pAxesWin;
  wxScrolledWindow*     m_pOverviewWin;

  wxMenu*               m_pAppearanceMenu;
  wxToolBar*            m_pToolbar;

  int                   m_direction;

  wxFileName            m_fnPointSet;

  void SetSimDirection(int dir);
  int GetSimDirection();

  TASIteratorThread*    m_pThreadIterator;
  wxCriticalSection     m_critSect;
  wxMutex               m_mutexIterate;
  wxCondition           m_conditionIterate;
  wxMutex               m_mutexWaiter;
  wxCondition           m_conditionWaiter;

private:
  wxMenu*               m_pFileMenu;
  wxMenu*               m_pOptionMenu;
  wxMenu*               m_pRefreshSubMenu;
  wxMenu*               m_pViewTypeSubMenu;
  wxTextCtrl*           m_pZoomCtrl;

  DECLARE_CLASS(TASSimulatorFrame)
  // any class wishing to process wxWindows events must use this macro
  DECLARE_EVENT_TABLE()
};


enum
{
  // menu items
  Menu_File_NewTAS = 2000,
  Menu_File_LoadTAS,
  Menu_File_ReloadTAS,
  Menu_File_SaveTAS,
  Menu_File_SaveTASAs,

  Menu_File_NewAssembly,
  Menu_File_LoadAssembly,
  Menu_File_ReloadAssembly,
  Menu_File_SaveAssembly,
  Menu_File_SaveAssemblyAs,

  Menu_file_SaveAssemblyImg,
  Menu_File_SavePointSet,

  Menu_File_Quit,

  Menu_Appearance_RefreshEvery = 2100,
  Menu_Appearance_RefreshScreenNow,
  Menu_Appearance_ClearDebugWindow,

  Menu_Edit_CopyTileType,
  Menu_Edit_FindCurrentTileInEditor,

  
  SubMenu_RefreshEvery_1 = 2150,
  SubMenu_RefreshEvery_2,
  SubMenu_RefreshEvery_3,
  SubMenu_RefreshEvery_4,
  SubMenu_RefreshEvery_5,
  SubMenu_RefreshEvery_10,
  SubMenu_RefreshEvery_20,
  SubMenu_RefreshEvery_50,
  SubMenu_RefreshEvery_75,
  SubMenu_RefreshEvery_100,
  SubMenu_RefreshEvery_CustomRate,
  SubMenu_RefreshEvery_Manual,

  Menu_Option_DimensionMode,
  Menu_Option_Dimension_2,
  Menu_Option_Dimension_3,
  Menu_Option_SimulationMode,
  Menu_Option_Simulation_aTAM,
  Menu_Option_Simulation_kTAM,
  Menu_Option_Simulation_2HAM,
  Menu_Option_SetTemperature,
  Menu_Option_SetStageFile,
  Menu_Option_UseTileConcentrations,
  Menu_Option_SetOneTileAdditions,
  Menu_Option_SetMaxTileAdditions,
  Menu_Option_SetUseTimingModel,
  Menu_Option_SetUse2HAMHack,
  Menu_Option_RandomSeed,
  Menu_Option_CheckLocalDeterminism,
  Menu_Option_ClearUserSettings,
  Menu_Option_DebugMsgs,
  Menu_Option_Debug,
  
  Menu_Control_Forward,
  Menu_Control_FastForward,
  Menu_Control_Stop,
  Menu_Control_Backward,
  Menu_Control_FastBackward,
  Menu_Control_SetBreakpoint,
  Menu_Control_BatchMode,
  Menu_Control_ResetAssembly,
  Menu_Control_TilesetIntoSimulator,
  Menu_Control_TilesetFromSimulator,
  Menu_Control_SetAsServer,
  Menu_Control_SetAsClient,
  
  Menu_Help_AboutProgram,
  Menu_Help_AboutLNSA,
  Menu_Help_AboutTileAssembly,
  Menu_Help_Help,
  
  ToolBar_Zoom_In,
  ToolBar_Zoom_Out,

  ToolBar_Modify_Frontier,

  ToolBar_TilesetIntoSimulator,
  ToolBar_TilesetFromSimulator,

  Menu_Appearance_AutoRotate,

  Menu_Appearance_ToggleUseEditor,
  Menu_Appearance_ToggleShowCellValues,
  Menu_Appearance_ToggleShowSpaceConfig,
  Menu_Appearance_ToggleShow2HAMConfig,
  Menu_Appearance_ToggleShowAxes,
  Menu_Appearance_ToggleShowOverview,
  Menu_Appearance_ToggleShowMessages,
  Menu_Appearance_ToggleShowMouseOver,
  Menu_Appearance_ToggleShowFrontier,
  Menu_Appearance_ToggleShowSymbols,
  Menu_Appearance_ToggleShowTileNames,
  Menu_Appearance_ToggleShowTileGlues,

  SubMenu_ViewType_Perspective,
  SubMenu_ViewType_Orthographic,
  Menu_Appearance_ViewType,
  Menu_Appearance_SetBackgroundColor,
  Menu_Appearance_SetTileColor,
  Menu_Appearance_SetTextColor,
  Menu_Appearance_ToggleAsyncDisplay,

  Menu_Appearance_ZoomIn,
  Menu_Appearance_ZoomOut,

  Menu_RightClick_TileAdd,

  Dlg_RandomSeed_Autogen,

  ID_TAS_RTCLICK_TTYPE,
  ID_TAS_POST_REDRAW,

  ID_TAS_SIMULATOR_FRAME,
  ID_TAS_EDITOR_FRAME
};

DECLARE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_INCR, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_INCR_WITH_SIG, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_FULL, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_FULL_WITH_SIG, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_2HAMUPDATE, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_2HAMPROGRESS, -1)


#endif // TASSimFrame_H
